/*
 * Licensed to Crate under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.  Crate licenses this file
 * to you under the Apache License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied.  See the License for the specific language governing
 * permissions and limitations under the License.
 *
 * However, if you have executed another commercial license agreement
 * with Crate these terms will supersede the license and you may use the
 * software solely pursuant to the terms of the relevant commercial
 * agreement.
 */

package io.crate.analyze;

import com.google.common.base.Function;
import com.google.common.base.Predicates;
import com.google.common.collect.Sets;
import io.crate.analyze.relations.JoinPair;
import io.crate.analyze.relations.QuerySplitter;
import io.crate.analyze.symbol.*;
import io.crate.metadata.Functions;
import io.crate.metadata.ReplaceMode;
import io.crate.metadata.ReplacingSymbolVisitor;
import io.crate.operation.operator.AndOperator;
import io.crate.planner.node.dql.join.JoinType;
import io.crate.sql.tree.QualifiedName;
import org.elasticsearch.common.collect.Tuple;

import javax.annotation.Nullable;
import java.util.*;

public class Rewriter {

    private final EvaluatingNormalizer normalizer;

    public Rewriter(Functions functions) {
        normalizer = EvaluatingNormalizer.functionOnlyNormalizer(functions, ReplaceMode.COPY);
    }

    /**
     * Rewrite an Outer join to an inner join if possible.
     * <p>
     * Conditions on OUTER tables are not pushed down when a MultiSourceSelect is initially created because
     * the whereClause needs to be checked AFTER the join
     * (because if the join generates NULL rows the whereClause could become TRUE on those NULL rows)
     * <p>
     * See the following two examples where <code>t2</code> is the OUTER table:
     * <p>
     * <pre>
     *     select * from t1
     *          left join t2 on t1.t2_id = t2.id
     *     where
     *          coalesce(t2.x, 20) > 10   # becomes TRUE for null rows
     * </pre>
     * <p>
     * <p>
     * <p>
     * But if we know that the whereClause cannot possible become TRUE then we can push it down
     * (re-writing this into an inner join)
     * <p>
     * This is possible because all rows that are generated by the left-join would be removed again by the whereClause anyway.
     * <p>
     * <pre>
     *     select * from t1
     *          left join t2 on t1.t2_id = t2.id
     *     where
     *          t2.x > 10   # if t2.x is NULL this is always FALSE
     * </pre>
     */
    public void tryRewriteOuterToInnerJoin(JoinPair joinPair,
                                           List<Symbol> mssOutputSymbols,
                                           QuerySpec multiSourceQuerySpec,
                                           QualifiedName left,
                                           QualifiedName right,
                                           QuerySpec leftQuerySpec,
                                           QuerySpec rightQuerySpec) {
        assert left.equals(joinPair.left()) : "This JoinPair has a different left Qualified name: " + joinPair.left();
        assert right.equals(joinPair.right()) :
            "This JoinPair has a different left Qualified name: " + joinPair.right();

        JoinType joinType = joinPair.joinType();
        if (!joinType.isOuter()) {
            return;
        }
        WhereClause where = multiSourceQuerySpec.where();
        if (!where.hasQuery()) {
            return;
        }
        final QualifiedName outerRelation;
        final QuerySpec outerSpec;
        if (joinType == JoinType.LEFT) {
            outerRelation = right;
            outerSpec = rightQuerySpec;
        } else {
            outerRelation = left;
            outerSpec = leftQuerySpec;
        }
        Map<Set<QualifiedName>, Symbol> splitQueries = QuerySplitter.split(where.query());
        Symbol outerRelationQuery = splitQueries.remove(Sets.newHashSet(outerRelation));
        if (outerRelationQuery == null) {
            return;
        }

        Symbol symbol = Symbols.replaceRelationColumn(
            outerRelationQuery,
            new Function<RelationColumn, Symbol>() {
                @Nullable
                @Override
                public Symbol apply(@Nullable RelationColumn input) {
                    if (input != null && input.relationName().equals(outerRelation)) {
                        return Literal.NULL;
                    }
                    return input;
                }
            }
        );
        Symbol normalized = normalizer.normalize(symbol, null);
        if (WhereClause.canMatch(normalized)) {
            return; // can't rewrite
        }

        applyOuterToInnerJoinRewrite(
            joinPair,
            mssOutputSymbols,
            multiSourceQuerySpec,
            outerSpec,
            outerRelation,
            splitQueries,
            outerRelationQuery
        );
    }

    private void applyOuterToInnerJoinRewrite(JoinPair joinPair,
                                              List<Symbol> mssOutputSymbols,
                                              QuerySpec multiSourceQuerySpec,
                                              QuerySpec outerSpec,
                                              QualifiedName outerRelation,
                                              Map<Set<QualifiedName>, Symbol> splitQueries,
                                              Symbol outerRelationQuery) {
        joinPair.joinType(JoinType.INNER);
        DereferenceRelationColumnFunction dereferenceRelationColumnFunction =
            new DereferenceRelationColumnFunction(outerRelation, outerSpec, mssOutputSymbols, joinPair.condition());
        outerSpec.where(outerSpec.where().add(Symbols.replaceRelationColumn(
            outerRelationQuery,
            dereferenceRelationColumnFunction)));
        if (splitQueries.isEmpty()) {
            multiSourceQuerySpec.where(WhereClause.MATCH_ALL);
        } else {
            multiSourceQuerySpec.where(new WhereClause(AndOperator.join(splitQueries.values())));
        }
        for (Tuple<RelationColumn, Symbol> symbolToRemove : dereferenceRelationColumnFunction.symbolsToNotCollect()) {
            final RelationColumn removedRC = symbolToRemove.v1();
            outerSpec.outputs().remove(symbolToRemove.v2());
            multiSourceQuerySpec.outputs().remove(removedRC);

            Function<Symbol, Symbol> replaceFunction = new RelationColumnIndexShifter(removedRC);
            joinPair.replaceCondition(replaceFunction);
            multiSourceQuerySpec.replace(replaceFunction);
        }
    }

    /**
     * Function to replace RelationColumns with the symbol they're pointing to
     * <p>
     * Also collects any RelationColumn which are being replaced and no longer required to be collected into {@link #symbolsToNotCollect()}
     */
    private static class DereferenceRelationColumnFunction implements Function<RelationColumn, Symbol> {
        private final QualifiedName outerRelation;
        private final QuerySpec outerSpec;
        private final List<Symbol> mssOutputSymbols;
        private final Symbol joinCondition;
        private final Set<Tuple<RelationColumn, Symbol>> symbolsToNotCollect;

        DereferenceRelationColumnFunction(QualifiedName outerRelation,
                                          QuerySpec outerSpec,
                                          List<Symbol> mssOutputSymbols,
                                          Symbol joinCondition) {
            this.outerRelation = outerRelation;
            this.outerSpec = outerSpec;
            this.mssOutputSymbols = mssOutputSymbols;
            this.joinCondition = joinCondition;
            this.symbolsToNotCollect = new HashSet<>();
        }

        @Nullable
        @Override
        public Symbol apply(@Nullable RelationColumn input) {
            if (input == null) {
                return null;
            }
            if (!input.relationName().equals(outerRelation)) {
                return input;
            }
            Symbol symbol = outerSpec.outputs().get(input.index());

            // if the column was only added to the outerSpec outputs because of the whereClause
            // it's possible to not collect it as long is it isn't used somewhere else
            if (!mssOutputSymbols.contains(symbol) &&
                !SymbolVisitors.any(Predicates.<Symbol>equalTo(input), joinCondition)) {
                symbolsToNotCollect.add(new Tuple<>(input, symbol));
            }
            return symbol;
        }

        Collection<Tuple<RelationColumn, Symbol>> symbolsToNotCollect() {
            return symbolsToNotCollect;
        }
    }

    /**
     * shift other relationColumns when one was removed
     * e.g. if the outerSpec had outputs: [o1, o2, o3]
     * and o2 got removed
     * a RelationColumn(o, 2) would now point to nothing instead of o3
     * so all indices needs to be decreased
     */
    private static class RelationColumnIndexShifter extends ReplacingSymbolVisitor<Void>
        implements Function<Symbol, Symbol> {

        private final RelationColumn removedRelationColumn;

        RelationColumnIndexShifter(RelationColumn removedRelationColumn) {
            super(ReplaceMode.COPY);
            this.removedRelationColumn = removedRelationColumn;
        }

        @Nullable
        @Override
        public Symbol apply(@Nullable Symbol input) {
            return process(input, null);
        }

        @Override
        public Symbol visitRelationColumn(RelationColumn relationColumn, Void context) {
            if (relationColumn.relationName().equals(removedRelationColumn.relationName())
                && relationColumn.index() > removedRelationColumn.index()) {

                return new RelationColumn(
                    relationColumn.relationName(),
                    relationColumn.index() - 1,
                    relationColumn.valueType());
            }
            return relationColumn;
        }
    }
}
